(************************************************************************)
(* constraints-sig.sml                                                  *)
(*                                                                      *)
(* Description:                                                         *)
(*    Boolean constraints used for the theorem prover.                  *)
(*                                                                      *)
(* Author:                                                              *)
(*    Evan Chang                                                        *)
(************************************************************************)

(* $Id: constraints-sig.sml,v 1.2 2002/05/14 21:33:02 bechang_conductor Exp $ *)


(************************************************************************)
(* signature CONSTRAINTS                                                *)
(*                                                                      *)
(*   Interface:                                                         *)
(*                                                                      *)
(*     Provides the required operations for using boolean constraints   *)
(*     to handle resourse non-determinism as described in [HP01].       *)
(*                                                                      *)
(************************************************************************)
signature CONSTRAINTS =
sig

  structure Prop  : PROP
  structure Cxt   : CXT


  (* Possible initialization code *)
  val init : unit -> unit


  (***********************)
  (* Boolean Expressions *)
  (***********************)
  type bexp

  val one : bexp
  val zero : bexp

  (* Check equality *)
  val bexpEq : bexp * bexp -> bool


  (************)
  (* Contexts *)
  (************)
  type linearCxt       = (Prop.prop * bexp) Cxt.Cxt
  type unrestrictedCxt = Prop.prop Cxt.Cxt

  val unresCxtToString : unrestrictedCxt -> string
  val linCxtToString   : linearCxt -> string


  (***************)
  (* Constraints *)
  (***************)
  type constraints

  (* Empty constraints *)
  val empty : constraints

  (* Merge constraints into a new set of constraints *)
  val addConstraints : constraints * constraints -> constraints

  (* Create constraint given boolean expression *)
  val constrainExp : bexp * bool -> constraints 

  (* See if the constraints are now unsatisfiable *)
  val unsatisfiable : constraints -> bool

  (*
     verifyEmpty (D,c) = NONE
     verifyEmpty (D,c) = SOME c'

     Verifies that a linear context D is empty.  A linear context
     is empty if each boolean expression is 0 or can be constrained to
     be zero.

     Returns:
       NONE    if unable to constrain D to "make" D empty.
       SOME c' if was able to constrain D;
               c' contains c plus new constraints to "make" D empty.
  *)
  val verifyEmpty : linearCxt * constraints -> constraints option

  (* distributeRes (D) = (DV,DV')

     Distribute resources in linear context D as described
     in [HP01].

     Returns:
       If   D   is {A1[e1], A2[e2], ..., An[en]}
       Then DV  is {A1[e1.x1], A2[e2.x2], ..., An[en.xn]}
       And  DV' is {A1[e1.x1'], A2[e2.x2'], ..., An[en.xn']}
         where x1,...,xn are new variables
         (x1' represents x1-bar)
   *)
  val distributeRes : linearCxt -> linearCxt * linearCxt
     
end; (* signature CONSTRAINTS *)
