(************************************************************************)
(* constraints.sml                                                      *)
(*                                                                      *)
(* Description:                                                         *)
(*    Boolean constraints used for the theorem prover.                  *)
(*                                                                      *)
(* Author:                                                              *)
(*    Evan Chang                                                        *)
(************************************************************************)

(* $Id: constraints.sml,v 1.2 2002/05/14 21:33:02 bechang_conductor Exp $ *)


(************************************************************************)
(* functor ConstraintsFn                                                *)
(*                                                                      *)
(* This implementation of the boolean contraints uses OBDDs to          *)
(* represent boolean expressions/constraints.                           *)
(*                                                                      *)
(************************************************************************)
functor ConstraintsFn(structure Prop' : PROP
                      structure Cxt' : CXT
                      structure Obdd' : OBDD
                      structure IDSrc' : UNIQUEIDSRC where type id = int)
:> CONSTRAINTS =
struct

  structure Prop = Prop'
  structure Cxt = Cxt'
  structure Obdd = Obdd'
  structure IDSrc = IDSrc'

  (* rnewVar : (unit -> int) ref

     Function to get a new variable id.
  *)
  val rnewVar : (unit -> int) ref = ref (fn () => ~1)

  (* init : unit -> unit *)
  fun init () = rnewVar := IDSrc.new ()


  (***********************)
  (* Boolean Expressions *)
  (***********************)
  type bexp = Obdd.obdd

  val one = Obdd.one
  val zero = Obdd.zero

  (* bexpEq : bexp * bexp -> bool *)
  val bexpEq = Obdd.eq


  (************)
  (* Contexts *)
  (************)
  type linearCxt       = (Prop.prop * bexp) Cxt.Cxt
  type unrestrictedCxt = Prop.prop Cxt.Cxt

  (* unresCxtToString : unrestrictedCxt -> string *)
  fun unresCxtToString (Cxt.Empty) = "."
    | unresCxtToString (G) =
      let
        fun unresCxtToString' (Cxt.Ext (Cxt.Empty, A)) = Prop.toString A
          | unresCxtToString' (Cxt.Ext (G',A)) =
              unresCxtToString' G' ^ ", " ^ Prop.toString A
      in
          unresCxtToString' G
        end

  (* linCxtToString : linearCxt -> string *)
  fun linCxtToString (Cxt.Empty) = "."
    | linCxtToString (D) =
      let
        fun linCxtToString' (Cxt.Ext (Cxt.Empty, (A,e))) = Prop.toString A
          | linCxtToString' (Cxt.Ext (D', (A,e))) =
              linCxtToString' D' ^ ", " ^ Prop.toString A
        in
          linCxtToString' D
        end


  (***************)
  (* Constraints *)
  (***************)
  type constraints = Obdd.obdd

  (* empty : constraints *)
  val empty = Obdd.one

  (* addConstraints : constraints * constraints -> constraints *)
  val addConstraints = Obdd.apply Obdd.And

  (* constrainExp : bexp * bool -> constraints *)
  fun constrainExp (b,true) = Obdd.apply Obdd.BiImp (b, Obdd.one)
    | constrainExp (b,false) = Obdd.apply Obdd.BiImp (b, Obdd.zero)

  (* unsatisfiable : constraints -> bool *)
  fun unsatisfiable c = Obdd.eq (c,zero)

  (* verifyEmpty : linearCxt * constraints -> constraints option

     verifyEmpty (D,c) = NONE
     verifyEmpty (D,c) = SOME c'

     Verifies that a linear context D is empty.  A linear context
     is empty if each boolean expression is 0 or can be constrained to
     be zero.

     Returns:
       NONE    if unable to constrain D to "make" D empty.
       SOME c' if was able to constrain D;
               c' contains c plus new constraints to "make" D empty.
  *)
  fun verifyEmpty (Cxt.Empty, c) = SOME c
    | verifyEmpty (Cxt.Ext (D, (A,e)), c) =
      if bexpEq (e,zero) then verifyEmpty (D,c)
      else if bexpEq (e,one) then NONE
      else
        let
          val c' = addConstraints (c, constrainExp (e,false))
        in
          if unsatisfiable c' then NONE
          else verifyEmpty (D,c')
        end

  (* distributeRes : linearCxt -> linearCxt * linearCxt
     
     distributeRes (D) = (DV,DV')

     Distribute resources in linear context D as described
     in Harland and Pym.

     Returns:
       If   D   is {A1[e1], A2[e2], ..., An[en]}
       Then DV  is {A1[e1.x1], A2[e2.x2], ..., An[en.xn]}
       And  DV' is {A1[e1.x1'], A2[e2.x2'], ..., An[en.xn']}
         where x1,...,xn are new variables
         (x1' represents x1-bar)
   *)
  fun distributeRes (Cxt.Empty) = (Cxt.Empty, Cxt.Empty)
    | distributeRes (Cxt.Ext (D, (A,e))) =
      let
        val x   = Obdd.mkVar (!rnewVar ())
        val ex  = addConstraints (e, x)
        val ex' = addConstraints (e, Obdd.not x)
        val (DV,DV') = distributeRes D
      in
        (Cxt.Ext (DV, (A,ex)), Cxt.Ext (DV', (A,ex')))
      end

end; (* functor ConstraintsFn *)
