(************************************************************************)
(* focusing-sig.sml                                                     *)
(*                                                                      *)
(* Description:                                                         *)
(*     Provides an inferface for the focusing phases.                   *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: focusing-sig.sml,v 1.1 2002/04/24 02:12:13 bechang_conductor Exp $ *)


(************************************************************************)
(* signature FOCUSING                                                   *)
(*                                                                      *)
(*   Interface:                                                         *)
(*                                                                      *)
(*     This signature aims to isolate each phase, so that one can       *)
(*     mix and match different implementations of the phases.           *)
(*     The "dispatcher" is the primary method for doing this.           *)
(*                                                                      *)
(************************************************************************)
signature FOCUSING =
sig

  structure C : CONSTRAINTS

  (* Failure continuation *)
  type ('a,'ans) fc = 'a -> 'ans

  (* Success continuation *)
  type ('a,'ans) sc = (C.constraints * 'a) * ('a,'ans) fc -> 'ans


  (******************)
  (* Sequents       *)
  (******************)
  datatype sequent =

    (* RightAsync (G,D,O,A)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       O (Omega) contains ordered linear hypotheses
       A is the goal
    *)
    RightAsync of C.unrestrictedCxt * C.linearCxt * C.linearCxt * C.Prop.prop

    (* LeftAsync (G,D,O,C)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       O (Omega) contains ordered linear hypotheses
       C is the goal that *may not be right asynchronous*
    *)
  | LeftAsync  of C.unrestrictedCxt * C.linearCxt * C.linearCxt * C.Prop.prop

    (* Decide (G,D,C)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       C is the goal that *may not be right asynchronous*
    *)
  | Decide of C.unrestrictedCxt * C.linearCxt * C.Prop.prop

    (* RightFocus (G,D,A)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       A is the goal being focused upon
    *)
  | RightFocus of C.unrestrictedCxt * C.linearCxt * C.Prop.prop

    (* LeftFocus (G,D,A,C)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       A is the hypothesis being focused upon
       C is the goal that *may not be right asynchronous*
    *)
  | LeftFocus  of C.unrestrictedCxt * C.linearCxt * C.Prop.prop * C.Prop.prop

  (* Give a string representation of a sequent *)
  val sequentToString : sequent -> string


  (******************)
  (* Proving Phases *)
  (******************)

  type ('a,'ans) dispatcher =
           sequent -> C.constraints * 'a -> ('a,'ans) sc * ('a,'ans) fc -> 'ans

  val rightAsync : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * C.linearCxt
                      * C.Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> 'ans


  val leftAsync  : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * C.linearCxt
                      * C.Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> 'ans


  val decide     : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * C.Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> 'ans


  val rightFocus : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * C.Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> 'ans


  val leftFocus  : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * C.Prop.prop
                      * C.Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> 'ans

end; (* signature FOCUSING *)
