(************************************************************************)
(* focusing.sml                                                         *)
(*                                                                      *)
(* Description:                                                         *)
(*     Implements the focusing phases in a straightforward sequential   *)
(*     manner.                                                          *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: focusing.sml,v 1.3 2002/06/21 04:12:32 bechang_conductor Exp $ *)


(************************************************************************)
(* functor Focusing                                                     *)
(*                                                                      *)
(* This implementation of the focusing phases is purely sequential      *)
(* using backtracking to handle disjunctive non-determinism.            *)
(*                                                                      *)
(* The parameter structure C' provides the ability to use boolean       *)
(* constraints for handling resource non-determinism.                   *)
(*                                                                      *)
(************************************************************************)
functor Focusing(structure C' : CONSTRAINTS)
:> FOCUSING =
struct

  structure C = C'
  structure Prop = C.Prop
  structure Cxt  = C.Cxt
  (*structure H = HeadCheck(structure Prop' = Prop)*)

  (* Failure continuation *)
  type ('a,'ans) fc = 'a -> 'ans

  (* Success continuation *)
  type ('a,'ans) sc = (C.constraints * 'a) * ('a,'ans) fc -> 'ans


  (* Add to context *)
  val ++ = Cxt.add
  infix ++

  (* Boolean Expressions *)
  val zero = C.zero
  val one = C.one


  (******************)
  (* Sequents       *)
  (******************)

  datatype sequent =

    (* RightAsync (G,D,O,A)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       O (Omega) contains ordered linear hypotheses
       A is the goal
    *)
    RightAsync of C.unrestrictedCxt * C.linearCxt * C.linearCxt * Prop.prop

    (* LeftAsync (G,D,O,C)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       O (Omega) contains ordered linear hypotheses
       C is the goal that *may not be right asynchronous*
    *)
  | LeftAsync  of C.unrestrictedCxt * C.linearCxt * C.linearCxt * Prop.prop

    (* Decide (G,D,C)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       C is the goal that *may not be right asynchronous*
    *)
  | Decide of C.unrestrictedCxt * C.linearCxt * Prop.prop

    (* RightFocus (G,D,A)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       A is the goal being focused upon
    *)
  | RightFocus of C.unrestrictedCxt * C.linearCxt * Prop.prop

    (* LeftFocus (G,D,A,C)
       
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       A is the hypothesis being focused upon
       C is the goal that *may not be right asynchronous*
    *)
  | LeftFocus  of C.unrestrictedCxt * C.linearCxt * Prop.prop * Prop.prop


  (* sequentToString : sequent -> string *)
  fun sequentToString (RightAsync (G,D,O,C)) =
        (C.unresCxtToString G ^ " ; " ^ C.linCxtToString D ^ " ; " ^
         C.linCxtToString O ^ "  ==>  " ^ Prop.toString C ^ " ^^")
    | sequentToString (LeftAsync  (G,D,O,C)) =
        (C.unresCxtToString G ^ " ; " ^ C.linCxtToString D ^ " ; " ^
         C.linCxtToString O ^ " ^^  ==>  " ^ Prop.toString C)
    | sequentToString (Decide     (G,D,C)) =
        (C.unresCxtToString G ^ " ; " ^ C.linCxtToString D ^
         " ; . ^^  ==>  " ^ Prop.toString C)
    | sequentToString (RightFocus (G,D,A)) =
        (C.unresCxtToString G ^ " ; " ^ C.linCxtToString D ^
         "  ==>  " ^ Prop.toString A ^ " vv")
    | sequentToString (LeftFocus  (G,D,A,C)) =
        (C.unresCxtToString G ^ " ; " ^ C.linCxtToString D ^ " ; " ^
         Prop.toString A ^ " vv  ==>  " ^ Prop.toString C)



  (******************)
  (* Proving Phases *)
  (******************)

  type ('a,'ans) dispatcher =
           sequent -> C.constraints * 'a -> ('a,'ans) sc * ('a,'ans) fc -> 'ans


  (* rightAsync : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * C.linearCxt * Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> 'ans

     rightAsync pfStep (G,D,O,A) (c,x) (sc,fc) = b

     Invariants:
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       O (Omega) contains ordered linear hypotheses
       A is the goal
       c are contraints that are not unsatisfiable

     Begin attempting to prove G;D;O ==> C by applying right
     invertible rules.
  *)
  fun rightAsync pfStep (G,D,O, Prop.Lolli(A,B)) d k =
        pfStep (RightAsync (G, D, O ++ (A,one), B)) d k
    | rightAsync pfStep (G,D,O, Prop.With(A,B)) d (sc,fc) =
      let
        (* Use lazy distribution of resources as defined in [HP01] *)
        val sc' = (fn (d',fc') =>
                     pfStep (RightAsync (G, D, O, B)) d' (sc,fc'))

      in
        (* Try to prove left branch first, then try the right branch
           *** Place for parallelism (AND-parallelism) *** *)
        pfStep (RightAsync (G, D, O, A)) d (sc',fc)
      end
    | rightAsync pfStep (G,D,O, Prop.Top) d (sc,fc) = 
        sc (d,fc)
    | rightAsync pfStep (G,D,O, Prop.Implies(A,B)) d k =
        pfStep (RightAsync (G ++ A, D, O, B)) d k
    | rightAsync pfStep (G,D,O, A) d (sc,fc) =
        (* C not right asynchronous, switch to applying left invertible
           rules *)
        pfStep (LeftAsync (G, D, O, A)) d (sc,fc)


  (* leftAsync  : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * C.linearCxt * Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> ans

     leftAsync pfStep (G,D,O,C) (c,x) (sc,fc) = b

     Invariants:
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       O (Omega) contains ordered linear hypotheses
       C is the goal that *may not be right asynchronus*
       c is not zero

     Begin attempting to prove G;D;O ==> C by applying left
     invertible rules.
  *)
  fun leftAsync pfStep (G,D, Cxt.Ext (O', (A, e)), C) (d as (c,x))
                (k as (sc,fc)) =
      let
        (* skip : unit -> 'ans *)
        fun skip () = pfStep (LeftAsync (G, D, O', C)) d k

        (* apply : Prop.prop * C.constraints -> 'ans *)
        fun apply (Prop.Tensor (A,B), c) =
              pfStep (LeftAsync (G, D, O' ++ (A,one) ++ (B,one), C)) (c,x) k
          | apply (Prop.Unit, c) = pfStep (LeftAsync (G, D, O', C)) (c,x) k
          | apply (Prop.Or (A,B), c) =
            let
              val sc' =
                 (fn (d',fc') =>
                   pfStep (LeftAsync (G, D, O' ++ (B,one), C)) d' (sc,fc'))
            in
              (* Again try to prove left branch first, then try the right
                 branch
                 *** Place for parallelism (AND-parallelism) *** *)
              pfStep (LeftAsync (G, D, O' ++ (A,one), C)) (c,x) (sc',fc)
            end
          | apply (Prop.Zero, c) = sc ((c,x),fc)
          | apply (Prop.Bang A, c) =
              pfStep (LeftAsync (G ++ A, D, O', C)) (c,x) k
          | apply (A, c) =
              (* A is not left asynchronous, put A into Delta *)
              pfStep (LeftAsync (G, D ++ (A,one), O', C)) (c,x) k
      in
        if C.bexpEq (e,zero) then skip ()
        else if C.bexpEq (e,one) then apply (A,c)
        else
          let
            val c' = C.addConstraints (c, C.constrainExp (e,true))
          in
            if C.unsatisfiable c' then skip ()
            else apply (A,c')
          end
      end
    | leftAsync pfStep (G,D, Cxt.Empty, C) d k =
        (* There are no more left asynchronous propositions in Omega, so
           we must now make a decision on what to focus on *)
        pfStep (Decide (G, D, C)) d k


  (* decide     : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> ans

     decide pfStep (G,D,C) (c,x) (sc,fc) = b

     Invariants:
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       C is the goal that *may not be right asynchronus*
       c is not zero

     Decide on a formula to focus on.
       + First try linear hypotheses
       + Then try unrestricted hypotheses
  *)
  fun decide pfStep (G, D, C) (c,x) (sc,fc) =
      let
        (* decideLBang : unrestrictedCxt -> 'a -> 'ans
           decideLBang (Gb) x = b

           Gb is the subset of G that has not been focused on
        
           Try focusing on each unrestricted hypothesis.  If we
           fail, then call the failure continuation.

           *** Place for parallelism -- OR-parallelism
        *)
        fun decideLBang (Cxt.Empty) x = fc x
          | decideLBang (Cxt.Ext (G',A)) x =
              pfStep (LeftFocus (G, D, A, C)) (c,x)
                     (sc, fn x' => decideLBang G' x')

        (* decideL : C.linearCxt * C.linearCxt -> 'a -> 'ans
           decideL (Db,Df) x = b

           Db is the subset of D that has not been been focused on
           Df is the subset of D not including the Db
        
           Try focusing on each linear hypothesis.  If we fail,
           then try focusing on unrestricted hypotheses.

           *** Place for parallelism -- OR-parallelism
        *)
        fun decideL (Cxt.Empty, _) x = decideLBang G x
          | decideL (Cxt.Ext (Db, (A,e)), Df) x =
            let
              (* skip : 'a -> 'ans *)
              fun skip x' = decideL (Db, Df ++ (A,e)) x'

              (* apply : C.constraints -> 'ans *)
              fun apply c =
                    (* TODO Possibly use a more efficient data structure for
                       linear contexts her *)
                    pfStep (LeftFocus (G, Cxt.combine (Db,Df), A, C)) (c,x)
                           (sc, skip)
            in
              if C.bexpEq (e,zero) then skip x
              else if C.bexpEq (e,one) then apply c
              else 
                let
                  val c' = C.addConstraints (c, C.constrainExp (e,true))
                in
                  if C.unsatisfiable c' then skip x
                  else apply c'
                end
            end
      in
        case C of
          (* C is atomic -- must try to focus on a hypothesis *)
          Prop.Atom P => decideL (D, Cxt.Empty) x
        | (* C is not atomic -- try to focus on the right first *)
          C => pfStep (RightFocus (G, D, C)) (c,x)
                      (sc, fn x' => decideL (D, Cxt.Empty) x')
      end


  (* rightFocus : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> 'ans

     rightFocus pfStep (G,D,A) (c,x) (sc,fc) = b

     Invariants:
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       A is the goal being focused upon
       c is not zero

     Apply right non-invertible rules to the focus formula.
  *)
  fun rightFocus pfStep (G,D, Prop.Tensor (A,B)) d (sc,fc) =
      let
        val (DV,DV') = C.distributeRes D
       
        (* Use lazy distribution of resources as defined in [HP01] *)
        val sc' = (fn (d',fc') => pfStep (RightFocus (G, DV', B)) d' (sc,fc'))
      in
        (* Try to prove left branch first, then try the right branch
           *** Place for parallelism (AND-parallelism) *** *)
        pfStep (RightFocus (G, DV, A)) d (sc',fc)
      end
    | rightFocus pfStep (G,D, Prop.Unit) (c,x) (sc,fc) =
      (case C.verifyEmpty (D,c) of
         NONE    => fc x
       | SOME c' => sc ((c',x),fc))
    | rightFocus pfStep (G,D, Prop.Or (A,B)) (c,x) (k as (sc,fc)) =
        (* *** Place for parallelism -- OR-parallelism *)
        pfStep (RightFocus (G, D, A)) (c,x)
               (sc, fn x' => pfStep (RightFocus (G, D, B)) (c,x') k)
    | rightFocus pfStep (G,D, Prop.Bang A) (c,x) (k as (sc,fc)) =
      (case C.verifyEmpty (D,c) of
         NONE    => fc x
       | SOME c' => pfStep (RightAsync (G, Cxt.Empty, Cxt.Empty, A)) (c',x) k)
    | rightFocus pfStep (G,D, A) d k =
        (* End focusing and switch back to applying right invertible rules *)
        pfStep (RightAsync (G, D, Cxt.Empty, A)) d k


  (* leftFocus  : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * Prop.prop * Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> 'ans

     leftFocus pfStep (G,D,A,C) (c,x) (sc,fc) = b
       
     Invariants:
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       A is the hypothesis being focused upon
       C is the goal that *may not be right asynchronous*
  *)
  fun leftFocus pfStep (G,D, A1 as Prop.Atom _, C) (d as (c,x)) (sc,fc) =
      (case C.verifyEmpty (D,c) of
         NONE    => fc x
       | SOME c' => 
           (case C of
              A2 as Prop.Atom _ => if Prop.eq (A1,A2) then sc ((c',x),fc)
                                   else fc x
            | C                 => fc x))
    | leftFocus pfStep (G,D, Prop.Lolli (A,B), C) d (sc,fc) =
      let
        val (DV,DV') = C.distributeRes D

        (* Try to prove the antecedent
             Use lazy distribution of resources as defined in [HP01] *)
        val sc' = (fn (d',fc') => pfStep (RightFocus (G, DV', A)) d' (sc,fc'))
      in
        (* Try to prove the succedent first, then try the antecedent
           *** Place for parallelism (AND-parallelism) *** *)
        pfStep (LeftFocus (G, DV, B, C)) d (sc',fc)
      end
    | leftFocus pfStep (G,D, Prop.With (A,B), C) (c,x) (k as (sc,fc)) =
        (* *** Place for parallelism -- OR-parallelism *)
        pfStep (LeftFocus (G, D, A, C)) (c,x)
               (sc, fn x' => pfStep (LeftFocus (G, D, B, C)) (c,x') k)
    | leftFocus pfStep (G,D, Prop.Implies (A,B), C) d (sc,fc) =
      let
        (* Try to prove the antecedent
             Use lazy distribution as defined in [HP01] *)
        val sc' = 
           (fn (d',fc') => pfStep (RightFocus (G, Cxt.Empty, A)) d' (sc,fc'))
      in
        (* Try to prove the succedent first, then try the antecedent
           *** Place for parallelism (AND-parallelism) *** *)
        pfStep (LeftFocus (G, D, B, C)) d (sc',fc)
      end
    | leftFocus pfStep (G,D, A, C) d k =
        (* A is not atomic and not left synchronous, so end focusing
           and switch back to applying left invertible rules *)
        pfStep (LeftAsync (G, D, Cxt.Ext(Cxt.Empty, (A,one)), C)) d k

end; (* functor Focusing *)
