(************************************************************************)
(* hashmapr.sml                                                         *)
(*                                                                      *)
(* Description:                                                         *)
(*    An implementation of a hash table.                                *)
(*                                                                      *)
(* Author:                                                              *)
(*    Evan Chang                                                        *)
(************************************************************************)

(* $Id: *)


(**************************************************************)
(* functor HashMapRFn                                         *)
(*                                                            *)
(* This implementation of a hash table uses separate chaining *)
(* to resolve collisions.  When the load factor               *)
(* (num_items / size) becomes 1.0, the table doubles, and     *)
(* when the load factor becomes 0.25, the table halves.       *)
(**************************************************************)
functor HashMapRFn(structure K : HASHABLE)
:> HASHMAPR where type key = K.key =
struct

  (******************************************)
  (* Tweakable Constants                    *)
  (******************************************)
  val MIN_SIZE = 32    (* should be a power of two *)


  (******************************************)
  (* Table                                  *)
  (******************************************)
  type key = K.key
  type 'a entry = key * 'a

  type hash = K.hash

  type 'a bucket = (hash * 'a entry) list
  type 'a table = 'a bucket array ref   (* the table *)
                  * int ref             (* number of entries *)
                  * int                 (* size hint *)

  (******************************************)
  (* Create a new table given a size hint   *)
  (******************************************)

  (* new : int -> 'a table *)
  fun new szHint =
      let
	(* Find the least power of two greater than szHint starting at m *)
        fun toPow2 m = if m >= szHint then m else toPow2 (2*m)
      in
	(ref (Array.array (toPow2 MIN_SIZE, nil)),
	 ref 0,
         szHint)
      end

  (* maybeResize : 'a table -> unit *)
  fun maybeResize (rtbl as ref tbl, ref n, szHint) =
      let
	(* resizeArr : 'a array * int -> 'a array *)
        fun resizeArr (arr, newSize) =
	    let
              val newarr = Array.array (newSize, nil)

              fun copybucket (nil) = ()
		| copybucket ((be as (h,e))::t) =
		  let
                    val i = (K.index h) mod newSize
                    val newbucket = Array.sub (newarr, i)
		  in
                    Array.update (newarr, i, be::newbucket);
		    copybucket t
		  end
	    in
              Array.app copybucket arr;
	      newarr
	    end

        val sz = Array.length tbl
        val szdiv2 = sz div 2
      in
        if n <= (sz div 4) andalso szdiv2 >= szHint then
	  rtbl := resizeArr (tbl, szdiv2)
        else if n >= sz then 
          rtbl := resizeArr (tbl, sz * 2)
        else
          ()
      end

  (* getIndex : 'a table -> key -> int * hash *)
  fun getIndex (ref tbl, _, _) key =
      let
        val sz = Array.length tbl
        val hash = K.hash key
      in
	((K.index hash) mod sz, hash)
      end

  (******************************************)
  (* Insert into the table                  *)
  (******************************************)

  (* removeBucket : key * 'a bucket * ('a bucket -> 'a bucket)
                        -> 'a entry option * 'a bucket *)
  fun removeBucket (_, nil, c) = (NONE, c nil)
    | removeBucket (key, (be as (_, e as (k,_)))::t, c) =
      if K.eq (key,k) then (SOME e, c t)
      else removeBucket (key, t, fn t' => c (be::t'))

  (* insertr : 'a table -> 'a entry -> 'a entry option *)
  fun insertr (t as (ref tbl, rn, _)) (e as (key, _)) =
      let
        val (i, hash) = getIndex t key
        val bucket = Array.sub (tbl, i)

        val (rem, newbucket) = removeBucket (key, bucket, fn x => x)

        val _ = Array.update (tbl, i, (hash,e)::newbucket)
        val _ = (case rem of
                   NONE   => (rn := !rn + 1; maybeResize t)
                 | SOME _ => ())
      in
        rem 
      end

  (* insert : 'a table -> 'a entry -> unit *)
  fun insert t e = (insertr t e; ())


  (******************************************)
  (* Lookup in the table                    *)
  (******************************************)

  (* lookup : 'a table -> key -> 'a option *)
  fun lookup (t as (ref tbl, _, _)) key =
      let
        val (i, _) = getIndex t key
        val bucket = Array.sub (tbl, i)

        val be = List.find (fn (_,(k,_)) => K.eq (key,k)) bucket
      in
        case be of
          NONE           => NONE
        | SOME (_,(_,v)) => SOME v
      end


  (******************************************)
  (* Remove Entries                         *)
  (******************************************)

  (* remove : 'a table -> key -> 'a entry option *)
  fun remove (t as (ref tbl, rn, _)) key = 
      let
        val sz = Array.length tbl
        val (i, _) = getIndex t key
        val bucket = Array.sub (tbl, i)

        val (rem, newbucket) = removeBucket (key, bucket, fn x => x)

        val _ = Array.update (tbl, i, newbucket)
        val _ = (case rem of
                   NONE   => ()
                 | SOME _ => (rn := !rn - 1; maybeResize t))

        val _ = maybeResize t
      in
        rem
      end


  (******************************************)
  (* Clear the table                        *)
  (******************************************)

  (* clear : 'a table -> unit *)
  fun clear (t as (ref tbl, rn, _)) =
      let in
	Array.modify (fn _ => nil) tbl;
        rn := 0;
        maybeResize t
      end


  (******************************************)
  (* Output List                            *)
  (******************************************)

  (* toList : 'a table -> 'a list *)
  fun toList (t as (ref tbl, _, _)) =
      let
        (* toListBucket : 'a bucket * 'a list -> 'a list *)
        fun toListBucket (b, acc) = foldr (fn ((_,(_,v)), a) => v::a) acc b
      in
        Array.foldr toListBucket nil tbl
      end

  (* toListi : 'a table -> 'a entry list *)
  fun toListi (t as (ref tbl, _, _)) =
      let
        (* toListBucket : 'a bucket * 'a list -> 'a list *)
        fun toListBucketi (b, acc) = foldr (fn ((_,e), a) => e::a) acc b
      in
        Array.foldr toListBucketi nil tbl
      end

end; (* functor HashMapRFn *)
