(************************************************************************)
(* headcheck.sml                                                        *)
(*                                                                      *)
(* Description:                                                         *)
(*     Implements an optimization for checking if a formula is worth    *)
(*     focusing on.                                                     *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: headcheck.sml,v 1.3 2002/06/21 04:12:32 bechang_conductor Exp $ *)

(************************************************************************)
(* functor HeadCheck                                                    *)
(*                                                                      *)
(*   Interface:                                                         *)
(*                                                                      *)
(*     This signature provides a check for determining if a formula     *)
(*     should be focused on as an optimization.                         *)
(*                                                                      *)
(*     Note, HeadCheck ascribes to HEADCHECK transparently to           *)
(*     side-step issues with structure sharing or the lack of           *)
(*     a "where structure" declaration.                                 *)
(*                                                                      *)
(************************************************************************)
functor HeadCheck(structure F' : FOCUSING) : HEADCHECK =
struct

  (* Focusing Structure *)
  structure F = F'

  (* Constraints Structure *)
  structure C = F.C

  (* Linear Logic Propositions *)
  structure P = F.C.Prop

  (* Contexts *)
  structure Cxt = F.C.Cxt

  (* Goals - parametric in P *)
  datatype goal =                       (* G ::=           *)
    GUnify of P.prop                    (*     | P' = P    *)
  | GTensor of goal * P.prop            (*     | G (x) A   *)
  | GOr of goal * goal                  (*     | G1 (+) G2 *)
  | GZero                               (*     | 0         *)
  | GTensorBang of goal * P.prop        (*     | G (x)! A  *)
  | GLolli of P.prop                    (*     | A -o P    *)


  (* precedence : goal -> int

     Assign precedence values to operators for printing.
  *)
  fun precedence (GZero)         = 10
    | precedence (GTensor _)     = 6
    | precedence (GTensorBang _) = 6
    | precedence (GOr _)         = 4
    | precedence (GLolli _)      = 2
    | precedence (GUnify _)      = 1 

  fun goalToString P C =
      let
        (* toString' : goal * int -> string
           toString' (C,n) = s

           n is the binding strength of the operator to the left
        *)
    fun toString' (C, n) =
            let
              fun parens C = "(" ^ toString' (C,0) ^ ")"
              val pr = precedence C
            in
              if n > pr then
                parens C
              else
                (case C of
                   GUnify (P')       => P.toString P' ^ " = " ^ P.toString P
                 | GTensor (G,A)     => toString' (G,pr) ^" (x) "^ P.toString A
                 | GTensorBang (G,A) => toString' (G,pr)^" (x)! "^ P.toString A
                 | GLolli A          => P.toString A ^" -o " ^ P.toString P
                 | GOr (G1,G2)       => toString' (G1,pr) ^" (+) "^
                                        toString' (G2,pr)
                 | GZero             => "0")
            end
      in
        toString' (C,0)
      end


  (* compile : P.prop -> goal
     compile A = G

     "Logically compile" focus formula A (parametric in goal P).
     ** TODO ** Memoize compilations of focus formulas.
  *)
  fun compile (P' as P.Atom _)  = GUnify P'
    | compile (P.Lolli (A,B))   = GTensor (compile B, A)
    | compile (P.With (A,B))    = GOr (compile A, compile B)
    | compile (P.Top)           = GZero
    | compile (P.Implies (A,B)) = GTensorBang (compile B, A)
    | compile (A)               = GLolli A 


  (* headcheck : (string -> unit)
                 -> F.C.unrestrictedCxt * F.C.linearCxt * P.prop * P.prop
                 -> C.constraints
                 -> (C.constraints * F.sequent list) option

     headcheck (G,D,A,P) = gs

     Given possible left focus formula A and atomic goal P.  Determine
     whether A possibly entails P.  If A possibly entails P, a list of
     remaining goals to be shown is returned.
  *)
  fun headcheck dbglog (G,D,A,P) c =
      let
        (* type fc = unit -> (C.constraints * F.sequent list) option
           type sc = (C.constraints * F.sequent list)
                       -> (C.constraints * F.sequent list) option

           headcheck' : C.unrestrictedCxt * C.linearCxt * goal
                        -> C.constraints
                        -> sc * fc
                        -> (C.constraints * F.sequent list) option
        *)
        fun headcheck' (G,D, GUnify P') c (sc,fc) =
	    (case C.verifyEmpty (D,c) of
	       NONE    => fc ()
	     | SOME c' => if P.eq (P',P) then sc (c',nil) else fc ())
	  | headcheck' (G,D, GTensor (g,A')) c (sc,fc) = 
            let
              val (DV,DV') = C.distributeRes D
              val residual = F.RightFocus (G,DV',A')
              val sc' = fn x => let val SOME (c',gs) = sc x
                                in  SOME (c',residual::gs)
                                end
            in
              step (G,DV,g) c (sc',fc)
	    end
	  | headcheck' (G,D, GOr (g1,g2)) c (sc,fc) = 
	    let
              val fc' = fn () => step (G,D,g2) c (sc,fc)
	    in
              step (G,D,g1) c (sc,fc')
	    end
	  | headcheck' (_,_, GZero) _ (_,fc) = fc ()
	  | headcheck' (G,D, GTensorBang (g,A')) c (sc,fc)= 
            let
              val residual = F.RightFocus (G, Cxt.Empty, A')
              val sc' = fn x => let val SOME (c',gs) = sc x
                                in  SOME (c',residual::gs)
                                end
	    in
              step (G,D,g) c (sc',fc)
	    end
	  | headcheck' (G,D, GLolli A') c (sc,fc) = 
              sc (c,[F.LeftAsync (G,D, Cxt.Ext(Cxt.Empty,(A',C.one)), P)])

        and step (G,D,goal) c k =
            let
              val _ = dbglog
                        ("Proving " ^ (C.unresCxtToString G) ^ " ; " ^
                         (C.linCxtToString D) ^ " ==> " ^
                         (goalToString P goal) ^ " ^")
            in
              headcheck' (G,D,goal) c k
            end

        val goal = compile A
      in
        step (G,D,goal) c (fn x => SOME x, fn () => NONE)
      end

end; (* functor HeadCheck *)
