(************************************************************************)
(* llprover-cc.sml                                                      *)
(*                                                                      *)
(* Description:                                                         *)
(*     An implementation of a theorem prover for linear logic for       *)
(*     the ConCert architecture.                                        *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: llprover-cc.sml,v 1.3 2002/06/21 04:12:32 bechang_conductor Exp $ *)


(************************************************************************)
(* functor LLProverCC                                                   *)
(*                                                                      *)
(* This implementation of a bottom-up theorem prover for linear logic   *)
(* is primarily based on the focusing sequent calculus presented in     *)
(* Pfenning's course notes [Pfe01] as the LLProver functor.             *)
(*                                                                      *)
(* This implementation attempts to extend the ideas learned from        *)
(* the CML implementation to utilize the ConCert architecture           *)
(* interface (see cctasks-sig.sml)                                      *)
(*                                                                      *)
(************************************************************************)
functor LLProverCC(structure Prop' : PROP
                   structure Cxt'  : CXT
                   structure Obdd' : OBDD) :> LLPROVER =
struct

  (* Constraints *)
  structure C' = ConstraintsFn(structure Prop' = Prop'
		               structure Cxt' = Cxt'
			       structure Obdd' = Obdd'
                               structure IDSrc' = IntUniqueIdSrcCML)

  (* Sequential Focusing Phases *)
  structure F = Focusing(structure C' = C')
  structure C = F.C

  (* Task Env *)
  datatype tosolve =
    Focusing of F.sequent
  | Resolve of F.sequent list

  (* Tasks *)
  structure Task =
  struct
    type env = tosolve * C.constraints
    datatype result = R of (C.constraints * ((env -> result) * env)) option
  end

  (* ConCert Interface *)
  structure CCTasks = CCTasksFn(structure T = Task)

  (* Propositions *)
  structure Prop = C.Prop

  (* Contexts *)
  structure Cxt = C.Cxt

  (* Head Check Optimization *)
  structure H = HeadCheck(structure F' = F)

  (* proveTask : Task.env ->  Tasks.result

     A proving task.  This function wraps everything together that is
     needed for a task to prove some subgoal.
  *)
  fun proveTask (seq, c) =
      let
        (* Add to context *)
        val ++ = Cxt.add
        infix ++

        (*********************)
        (* Logging Utilities *)
        (*********************)

        (* log : Logger.level -> string -> unit *)
        fun log lvl msg =
            let
              val str = (CML.tidToString (CML.getTid ())) ^ ": " ^ msg ^ "\n" 
            in
              Logger.log lvl str
            end

        val dbgMsg = log Logger.DEBUG
        val noteMsg = log Logger.NOTE
        val headcheck = H.headcheck dbgMsg


        (******************)
        (* Proving Phases *)
        (******************)
      
        (* decide : ('a,'ans) F.dispatcher
                    -> C.unrestrictedCxt * C.linearCxt * Prop.prop
                    -> C.constraints * unit
                    -> ('a,'ans) F.sc * ('a,'ans) F.fc
                    -> 'ans
      
           decide (G,D,C) (c,()) (sc,fc) = b
      
           Invariants:
             G (Gamma) contains unrestricted hypotheses
             D (Delta) contains linear hypotheses that *may not be left
                       asynchronous*
             C is the goal that *may not be right asynchronus*
             c is not zero
      
           Decide on a formula to focus on.
             + First try linear hypotheses
             + Then try unrestricted hypotheses
        *)
        fun decide pfStep (G, D, C) (c,()) (sc,fc) =
            let
              (* decideLBang : C.unrestrictedCxt -> task list -> task list
                 decideLBang Gb tsks = tsks'
      
                 Gb is the subset of G that has not been focused on
      
                 tsks is the list of tasks for each focus formula.
              
                 Start a task to focus on each unrestricted hypothesis. 
              *)
              fun decideLBang (Cxt.Empty) tsks = tsks
                | decideLBang (Cxt.Ext (G',A)) tsks =
                  let
		    open CCTasks
      
                    fun noheadcheck () =
			let
			  (* Start the child task *)
			  val s = Focusing (F.LeftFocus (G, D, A, C))
			  val task = inject true (proveTask, (s,c))
			in
			  (* Continue to the next formula *)
			  decideLBang G' (task::tsks)
			end
                  in
                    if Constants.USE_HEADCHECK then
                      (case C of
                         Prop.Atom _ =>
                           (case headcheck (G,D,A,C) c of
			      NONE => decideLBang G' tsks
			    | SOME (c',gs) =>
			      let
                                val s = Resolve gs
				val task = inject true (proveTask, (s,c'))
			      in
				decideLBang G' (task::tsks)
			      end)
		       | _ => noheadcheck ())
		    else noheadcheck ()
                  end
      
      
              (* decideL : C.linearCxt * C.linearCxt -> task list -> task list
                 decideL (Db,Df, tsks) = tsks'
      
                 Db is the subset of D that has not been been focused on
                 Df is the subset of D not including the Db
      
                 tsks is the list of tasks for each focus formula.
                 
                 Start a task to focus on each linear hypothesis. 
              *)
              fun decideL (Cxt.Empty, _) tsks = decideLBang G tsks
                | decideL (Cxt.Ext (Db, (A,e)), Df) tsks =
                  let
                    (* skip : unit -> (task * int) list *)
                    fun skip () = decideL (Db, Df ++ (A,e)) tsks
      
                    (* apply : C.constraints -> (task * int) list *)
                    fun apply c =
                        let
                          open CCTasks
      
                          fun noheadcheck () =
			      let
				(* TODO Possibly use a more efficient data
                                   structure for linear contexts here *)
				val s = Focusing (F.LeftFocus
					  (G,Cxt.combine (Db,Df), A, C))
				val task = inject true (proveTask, (s,c))
                              in
                                (* Continue to the next formula *)
				  decideL (Db, Df ++ (A,e)) (task::tsks)
			      end
			in
			  if Constants.USE_HEADCHECK then
                            (case C of
			       Prop.Atom _ => 
			        (case headcheck (G,Cxt.combine(Db,Df),A,C) c of
				   NONE => skip ()
				 | SOME (c',gs) =>
				   let
                                     val s = Resolve gs
				     val task = inject true (proveTask, (s,c'))
				   in
				     decideL (Db, Df ++ (A,e)) (task::tsks)
				   end)
			     | _ => noheadcheck ())
			  else noheadcheck ()
                        end
                  in
                    if C.bexpEq (e,C.zero) then skip ()
                    else if C.bexpEq (e,C.one) then apply c
                    else 
                      let
                        val c' = C.addConstraints (c, C.constrainExp (e,true))
                      in
                        if C.unsatisfiable c' then skip ()
                        else apply c'
                      end
                  end
      

              (* resultLoop : task list -> 'ans *)
              fun resultLoop (nil) = fc ()
                | resultLoop (ts)  =
                  let
                    open CCTasks
      
                    val _ = dbgMsg "Waiting for results"
                    val (Task.R result, ts') = relax ts
                    val _ = dbgMsg "  Received result"
                  in
                    case result of
                      NONE => resultLoop ts'
                    | SOME (c,next) =>
      		      let
      			fun fc' _ =
      			    let
			      val t = inject true next
      			    in
      			      resultLoop (t::ts')
      			    end
      		      in
                        (* *FIXME* cheating a little for now because the
			   code for fc' is not closed *)
      			sc ((c,()), fc')
      		      end
                  end
            in
              case C of
                (* C is atomic -- must try to focus on a hypothesis *)
                Prop.Atom _ =>
		  let
                    val tsks = decideL (D, Cxt.Empty) nil
		  in
                    resultLoop tsks
		  end
              | (* C is not atomic -- try to focus on the right first *)
                C =>
		  let
                    open CCTasks

                    (* Start the child task *)
                    val sequent = Focusing (F.RightFocus (G, D, C))
                    val task = inject true (proveTask, (sequent,c))
                    val tsks = decideL (D, Cxt.Empty) [task]
		  in
                    resultLoop tsks
		  end
            end
      
      
        (* pfStep : F.sequent
                    -> C.constraints * 'a
                    -> 'a F.sc * 'a F.fc
                    -> F.ans
      
           pfStep seq (c,) (sc,fc) = b
      
           Make a step in the proof search.
        *)
        fun pfStep seq d k =
            let
              fun doit (F.RightAsync (s)) = F.rightAsync pfStep s d k
                | doit (F.LeftAsync  (s)) = F.leftAsync pfStep s d k
                | doit (F.Decide     (s)) = decide pfStep s d k
                | doit (F.RightFocus (s)) = F.rightFocus pfStep s d k
                | doit (F.LeftFocus  (s)) = F.leftFocus pfStep s d k
      
              val _ = dbgMsg ("Proving " ^ (F.sequentToString seq))
            in
              doit seq
            end


        (* *FIXME* cheating a little for now because the code for fc' is not
           closed *)
        fun sc' ((c',()), fc') =
              Task.R (SOME (c', (fn _ => fc' (), (seq,c))))
      in
        (case seq of
          Focusing s => pfStep s (c,()) (sc', fn _ => Task.R NONE)
        | Resolve gs => 
          let
            fun resl d (nil) = sc' d
	      | resl (d as ((c',()),fc)) (h::t) =
                  pfStep h (c',()) (fn d' => resl d' t, fc)
	  in
            resl ((c,()), fn _ => Task.R NONE) gs
	  end)
      end
      

  (*********************)
  (* Proving Interface *)
  (*********************)

  (* prove : Prop.prop Cxt.Cxt * Prop.prop Cxt.Cxt * Prop.prop -> bool *)
  fun prove (G, D, C) =
      let
        open CCTasks 

        fun mkInitCxt (Cxt.Empty) = Cxt.Empty
          | mkInitCxt (Cxt.Ext (D', A)) = Cxt.Ext (mkInitCxt D', (A,C.one))

        fun initTask x = (C.init (); proveTask x)

        val sequent = Focusing (F.RightAsync (G, Cxt.Empty, mkInitCxt D, C))
        val (Task.R r) = startSim 0.0 (initTask, (sequent,C.empty))
      in
        case r of
          NONE => false
        | SOME _ => true
      end


end; (* functor LLProverCML *)
