(************************************************************************)
(* llprover-cml.sml                                                     *)
(*                                                                      *)
(* Description:                                                         *)
(*     An implementation of a theorem prover for linear logic using     *)
(*     CML to explore parallelzing such a theorem prover.               *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: llprover-cml.sml,v 1.3 2002/06/21 04:12:32 bechang_conductor Exp $ *)


(************************************************************************)
(* functor LLProverCML                                                  *)
(*                                                                      *)
(* This implementation of a bottom-up theorem prover for linear logic   *)
(* is primarily based on the focusing sequent calculus presented in     *)
(* Pfenning's course notes [Pfe01] as the LLProver functor.             *)
(*                                                                      *)
(* This implementation uses concurrency to handle some disjunctive      *)
(* non-determinism by trying all the possiblities at once.  Though      *)
(* this could be applied in many locations, this version only           *)
(* uses parallelization at the decision phase.                          *)
(*                                                                      *)
(************************************************************************)
functor LLProverCML(structure Prop' : PROP
                    structure Cxt'  : CXT
                    structure Obdd' : OBDD) :> LLPROVER =
struct

  (* Constraints *)
  structure C' = ConstraintsFn(structure Prop' = Prop'
		               structure Cxt' = Cxt'
			       structure Obdd' = Obdd'
                               structure IDSrc' = IntUniqueIdSrcCML)

  (* Sequential Focusing Phases *)
  structure F = Focusing(structure C' = C')
  structure C = F.C
     
  (* Propositions *)
  structure Prop = F.C.Prop

  (* Contexts *)
  structure Cxt  = F.C.Cxt

  (* Head Check Optimization *)
  structure H = HeadCheck(structure F' = F)

  (* Add to context *)
  val ++ = Cxt.add
  infix ++


  (*********************)
  (* Tunable Constants *)
  (*********************)

  (* NSTEPS_POLL_STOP : int

     Number of steps of proving to execute without checking if we have been
     signaled to die (i.e. 0 means check each step).
  *)
  val NSTEPS_POLL_STOP = 3


  (*********************)
  (* Logging Utilities *)
  (*********************)

  (* log : Logger.level -> string -> unit *)
  fun log lvl msg =
      let
        val str = (CML.tidToString (CML.getTid ())) ^ ": " ^ msg ^ "\n" 
      in
        Logger.log lvl str
      end

  val dbgMsg = log Logger.DEBUG
  val noteMsg = log Logger.NOTE
  val headcheck = H.headcheck dbgMsg


  (*****************)
  (* Communication *)
  (*****************)

  datatype Message = 
    Failure of CML.thread_id
  | Success of C.constraints
  | STOP
  | NEXT

  exception UnexpectedMessage

  (* MessageToString : Message -> string *)
  fun MessageToString (Failure tid) = "Failure (" ^ (CML.tidToString tid) ^ ")"
    | MessageToString (Success _)   = "Success"
    | MessageToString (STOP)        = "STOP"
    | MessageToString (NEXT)        = "NEXT"
   

  (* Communication *)
  type commLink = { stopIn    : Message SyncVar.ivar,
                    nextIn    : Message Multicast.port,
                    stopOut   : Message SyncVar.ivar list,
                    nextOut   : Message Multicast.mchan,
                    resultOut : Message CML.chan }

  (* killChildren : commLink -> unit *)
  fun killChildren ({stopOut, ...} : commLink) = 
        List.app (fn iv => SyncVar.iPut (iv, STOP)) stopOut

  (* die : commLink -> 'a

     Signal my children to die and exit.
  *)
  fun die comm =
      let
        val _ = killChildren comm
        val _ = noteMsg "Exiting"
      in
        CML.exit ()
      end

  (* step : ('a -> 'b) -> 'a -> commLink -> 'b
  
     step f x = y

     Possibly check to see if we have been signaled to die (based on
     NSTEPS_POLL_STOP), then make a step in computation by applying
     function f to argument x.
  *)
  local
    val stepcnt = ref 0
  in
    fun step f x (comm as {stopIn, ...} : commLink) =
        let
           (* maybeDie : unit -> unit
  
              Poll to see if I have been signaled to die.  If so, signal my
              children to die and exit.
           *)
           fun maybeDie () =
               (case (SyncVar.iGetPoll stopIn) of
                  NONE      => ()
                | SOME STOP => die comm
                | _         => raise UnexpectedMessage)
        in
          if (!stepcnt)=NSTEPS_POLL_STOP then
            (stepcnt := 0; maybeDie (); f x)
          else
            (stepcnt := !stepcnt + 1; f x)
        end
  end


  (******************)
  (* Proving Phases *)
  (******************)

  (* decide : ('a,'ans) F.dispatcher
              -> C.unrestrictedCxt * C.linearCxt * Prop.prop
              -> C.constraints * commLink
              -> (commLink,'ans) F.sc * (commLink,'ans) F.fc
              -> 'ans

     decide (G,D,C) (c,comm) (sc,fc) = b

     Invariants:
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       C is the goal that *may not be right asynchronus*
       c is not zero

     Decide on a formula to focus on.
       + First try linear hypotheses
       + Then try unrestricted hypotheses
  *)
  fun decide pfStep (G, D, C) (c, comm as {nextOut, ...} : commLink) (sc,fc) =
      let
        (* respondSuccess : (C.constraints * commLink) * (commLink,'ans) F.fc
                            -> 'ans
        
           Respond to the parent thread that we succeeded.  After
           responding, wait for instructions from the parent.
        *)
        fun respondSuccess ((c',
                             comm as {stopIn,nextIn,resultOut,...} : commLink),
                            fc') =
            let
              open CML

              val _ = noteMsg "  Responding with Success"
              val reply = select [
                            wrap (sendEvt (resultOut, Success c'),
                                  (fn () => select [
                                              SyncVar.iGetEvt   stopIn,
                                              Multicast.recvEvt nextIn
                                            ])),
                            SyncVar.iGetEvt stopIn
                          ]
              val _ = noteMsg ("    Received message " ^
                               (MessageToString reply))
            in
              (case reply of
                 STOP => die comm 
               | NEXT => fc' comm
               | _    => raise UnexpectedMessage)
            end

        (* respondFailure : commLink -> 'ans

           Respond to the parent thread that we failed.  Just send
           the message and return.
        *)
        fun respondFailure (comm as {resultOut, ...} : commLink) =
            let
              open CML
              val _ = noteMsg "  Responding with Failure"
            in
              (* TODO Find a way to die immediately rather than
                      waiting for synchronization. *)
              (* (sendEvt (resultOut, Failure (getTid())); die comm; false) *)
              send (resultOut, Failure (getTid()));
              die comm;
              false
            end

        (* addStopOut : commLink * Command SyncVar.ivar -> commLink
           
           Add an ivar to the commLink.
        *)
        fun addStopOut ({stopIn, nextIn, stopOut,
                         nextOut, resultOut}: commLink, iv) =
            let
              val newComm = { stopIn    = stopIn,
                              nextIn    = nextIn,
                              stopOut   = iv::stopOut,
                              nextOut   = nextOut,
                              resultOut = resultOut }
            in
              newComm
            end

        (* decideLBang : C.unrestrictedCxt 
                         -> commLink * (Message event * CML.thread_id) list
                         -> commLink * (Message event * CML.thread_id) list
           decideLBang (Gb, evts) = evts'

           Gb is the subset of G that has not been focused on

           evts is the list of events to listen for responses from child
                   threads paired with the child thread id.
        
           Spawn a thread to focus on each unrestricted hypothesis. 
        *)
        fun decideLBang (Cxt.Empty) (comm,evts) = (comm,evts)
          | decideLBang (Cxt.Ext (G',A)) (comm,evts) =
	    let
	      open CML

	      (* Create new commLink for the child thread *)
	      val childComm = { stopIn    = SyncVar.iVar (),
                                nextIn    = Multicast.port nextOut,
                                stopOut   = nil,
                                nextOut   = Multicast.mChannel (),
                                resultOut = channel () }
  
	      (* Add the stop ivar for the child to the list managed by
                 the parent *)
	      val newComm = addStopOut (comm, #stopIn childComm)
  
	      fun noheadcheck () = 
		  let
		    (* Spawn the child thread *)
		    val tid =
			spawn (fn () =>
			  (pfStep (F.LeftFocus (G, D, A, C))
				  (c, childComm)
				  (respondSuccess, respondFailure);()))
		  in
		    step (decideLBang G')
			 (newComm, (recvEvt (#resultOut childComm), tid)::evts)
			 newComm
		  end
	    in
	      if Constants.USE_HEADCHECK then
		(case C of
		   Prop.Atom _ =>
                   (case headcheck (G,D,A,C) c of
		      NONE => decideLBang G' (comm, evts)
                    | SOME (c',gs) =>
                      let
                        fun resl d (nil) = respondSuccess d
			  | resl (d as ((c',cm'),fc)) (h::t) =
                              pfStep h (c',cm') (fn d' => resl d' t, fc)

			val tid =
			    spawn (fn () =>
			      (resl ((c',childComm), respondFailure) gs;
                               ()))
		      in
			  step (decideLBang G')
			       (newComm,
				(recvEvt (#resultOut childComm), tid)::evts)
			       newComm
		      end)
		 | _ => noheadcheck ())
	      else noheadcheck ()
	    end


        (* decideL : C.linearCxt * C.linearCxt
                     -> commLink * (Message event * CML.thread_id) list
                     -> commLink * (Message event * CML.thread_id) list
           decideL (Db,Df, evts) = evts'

           Db is the subset of D that has not been been focused on
           Df is the subset of D not including the Db

           evts is the list of events to listen for responses from child
                   threads paired with the child thread id.
           
           Spawn a thread to focus on each linear hypothesis. 
        *)
        fun decideL (Cxt.Empty, _) (comm,evts) = decideLBang G (comm,evts)
          | decideL (Cxt.Ext (Db, (A,e)), Df) (comm,evts) =
            let
              (* skip : unit
                        -> commLink * Message event * CML.thread_id list *)
              fun skip () = decideL (Db, Df ++ (A,e)) (comm,evts)

              (* apply : C.constraints
                         -> commLink * Message event * CML.thread_id list *)
              fun apply c =
                  let
                    open CML
  
                    (* Create new commLink for the child thread *)
                    val childComm = { stopIn    = SyncVar.iVar (),
                                      nextIn    = Multicast.port nextOut,
                                      stopOut   = nil,
                                      nextOut   = Multicast.mChannel (),
                                      resultOut = channel () }
  
                    (* Add the stop ivar for the child to the list managed by
                       the parent *)
                    val newComm = addStopOut (comm, #stopIn childComm)
  
                    fun noheadcheck () =
			let
			  (* TODO Possibly use a more efficient data structure
                             for linear contexts here *)
			  val tid =
			      spawn (fn () =>
			      (pfStep (F.LeftFocus (G,Cxt.combine(Db,Df),A,C))
                                      (c, childComm)
                                      (respondSuccess, respondFailure); ()))
			in
			  step (decideL (Db, Df ++ (A,e)))
                               (newComm,
				(recvEvt (#resultOut childComm), tid)::evts)
                               newComm
			end
		  in
		    if Constants.USE_HEADCHECK then
		      (case C of
			 Prop.Atom _ => 
			 (case headcheck (G,Cxt.combine(Db,Df),A,C) c of
			      NONE => skip ()
			    | SOME (c',gs) =>
			      let
				fun resl d (nil) = respondSuccess d
				  | resl (d as ((c',cm'),fc)) (h::t) =
				      pfStep h (c',cm')
					     (fn d' => resl d' t, fc)
				      
				val tid =
				    spawn (fn () =>
				     (resl ((c',childComm), respondFailure) gs;
				      ()))
			      in
				step (decideL (Db, Df ++ (A,e)))
				     (newComm,
			              (recvEvt (#resultOut childComm),
				       tid)::evts)
				     newComm
			      end)
		       | _ => noheadcheck ())
		    else noheadcheck ()
		  end
            in
              if C.bexpEq (e,C.zero) then skip ()
              else if C.bexpEq (e,C.one) then apply c
              else 
                let
                  val c' = C.addConstraints (c, C.constrainExp (e,true))
                in
                  if C.unsatisfiable c' then skip ()
                  else apply c'
                end
            end


        (* resultLoop : commLink * (Message event * CML.thread_id list) *)
        fun resultLoop (comm, nil) = fc comm
          | resultLoop (comm, es)  =
            let
              open CML

              (* If asked, ask for the next answer from my children.
		 Then restart select from the children that are still
                 alive.
              *)
              fun askNext (newComm : commLink) =
                  let
                    val _ = Multicast.multicast (#nextOut newComm, NEXT)
                  in
                    resultLoop (newComm, es)
                  end

              (* Also listen for signal from my parent *)
              val _ = noteMsg "Waiting for results"
              val result = select (SyncVar.iGetEvt (#stopIn comm) ::
                                   (#1 (ListPair.unzip es)))
              val _ = noteMsg ("  Received message " ^
                               (MessageToString result))
            in
              (case result of
                 Failure failTid =>
                   resultLoop (comm,
                               (List.filter (fn (_,t) => failTid <> t) es))
               | Success c' => sc ((c',comm), askNext)
               | STOP => die comm
               | _ => raise UnexpectedMessage)
            end
      in
        case C of
          (* C is atomic -- must try to focus on a hypothesis *)
          Prop.Atom _ =>
            let
              val (newComm, evts) = decideL (D, Cxt.Empty) (comm,nil)
	    in
              resultLoop (newComm, evts)
	    end
        | (* C is not atomic -- also try to focus on the right *)
          C =>
            let
              open CML

              (* Create new commLink for the child thread *)
              val childComm = { stopIn    = SyncVar.iVar (),
                                nextIn    = Multicast.port nextOut,
                                stopOut   = nil,
                                nextOut   = Multicast.mChannel (),
                                resultOut = channel () }
  
              (* Add the stop ivar for the child to the list managed by
                 the parent *)
              val newComm = addStopOut (comm, #stopIn childComm)
  
              (* TODO Possibly use a more efficient data structure for
	         linear contexts here *)
              val tid = spawn (fn () =>
                          (pfStep (F.RightFocus (G, D, C))
                                  (c, childComm)
                                  (respondSuccess, respondFailure); ()))

              val (newComm, evts) =
                decideL (D, Cxt.Empty)
                        (newComm, [(recvEvt (#resultOut childComm), tid)])
            in
              resultLoop (newComm, evts)
	    end
      end


  (* pfStep : F.sequent
              -> C.constraints * 'a
              -> ('a,'ans) F.sc * ('a,'ans) F.fc
              -> 'ans

     pfStep seq (c,comm) (sc,fc) = b

     Make a step in the proof search.
  *)
  fun pfStep seq (d as (_,comm)) k =
      let
        fun doit (F.RightAsync (s)) = F.rightAsync pfStep s d k
          | doit (F.LeftAsync  (s)) = F.leftAsync pfStep s d k
          | doit (F.Decide     (s)) = decide pfStep s d k
          | doit (F.RightFocus (s)) = F.rightFocus pfStep s d k
          | doit (F.LeftFocus  (s)) = F.leftFocus pfStep s d k

        val _ = dbgMsg ("Proving " ^ (F.sequentToString seq))
      in
        step doit seq comm
      end


  (*********************)
  (* Proving Interface *)
  (*********************)

  (* prove : Prop.prop Cxt.Cxt * Prop.prop Cxt.Cxt * Prop.prop -> bool *)
  fun prove (G, D, C) =
      let
        fun prove' (G, D, C) =
            let
              val _ = C.init ()

              fun mkInitCxt (Cxt.Empty) = Cxt.Empty
                | mkInitCxt (Cxt.Ext (D', A)) =
                    Cxt.Ext (mkInitCxt D', (A,C.one))
      
              val comm = { stopIn    = SyncVar.iVar (),
                           nextIn    = Multicast.port (Multicast.mChannel ()),
                           stopOut   = nil,
                           nextOut   = Multicast.mChannel (),
                           resultOut = CML.channel () }
      
            in
              pfStep (F.RightAsync (G, Cxt.Empty, mkInitCxt D, C))
                     (C.empty, comm)
                     (fn ((_,cm),_) => (killChildren cm; true),
                      fn cm => (killChildren cm; false))
            end

        val result = ref false
        val _ = RunCML.doit ((fn () => result := prove' (G, D, C)), NONE)
      in
        !result
      end

end; (* functor LLProverCML *)
