(************************************************************************)
(* llprover.sml                                                         *)
(*                                                                      *)
(* Description:                                                         *)
(*     A straightforward implementation of a theorem prover for linear  *)
(*     logic.  See discussion below.                                    *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: llprover.sml,v 1.6 2002/06/21 04:12:32 bechang_conductor Exp $ *)

(************************************************************************)
(* functor LLProver                                                     *)
(*                                                                      *)
(* This implementation of a bottom-up theorem prover for linear logic   *)
(* is primarily based on the focusing sequent calculus presented in     *)
(* Pfenning's course notes [Pfe01].  The implementation of the          *)
(* of the focusing phases is in the Focusing functor (focusing-sig.sml/ *)
(* focusing.sml).  To deal with resource non-determinism, we use        *)
(* the boolean constraints idea discussed in [HP01] using OBDDs to      *)
(* implement the constraints (see constraints-sig.sml).                 *)
(*                                                                      *)
(* This implementation is purely sequential, so disjunctive             *)
(* non-determinism is handled by backtracking.                          *)
(*                                                                      *)
(************************************************************************)
functor LLProver(structure Prop' : PROP
                 structure Cxt'  : CXT
                 structure Obdd' : OBDD) :> LLPROVER =
struct

  (* Constraints *)
  structure C' = ConstraintsFn(structure Prop' = Prop'
                               structure Cxt' = Cxt'
                               structure Obdd' = Obdd'
		 	       structure IDSrc' = IntUniqueIdSrc)

  (* Sequential Focusing Phases *)
  structure F = Focusing(structure C' = C')
  structure C = F.C
  structure H = HeadCheck(structure F' = F)

  (* Propositions *)
  structure Prop = F.C.Prop

  (* Contexts *)
  structure Cxt = F.C.Cxt


  (*********************)
  (* Logging Utilities *)
  (*********************)

  (* log : Logger.level -> string -> unit *)
  fun log lvl msg =
      let
        val str = msg ^ "\n" 
      in
        Logger.log lvl str
      end    

  val dbgMsg = log Logger.DEBUG
  val noteMsg = log Logger.NOTE
  val headcheck = H.headcheck dbgMsg

  (* Add to context *)
  val ++ = Cxt.add
  infix ++

  (******************)
  (* Proving Phases *)
  (******************)

  (* decide     : ('a,'ans) dispatcher
                   -> C.unrestrictedCxt * C.linearCxt * Prop.prop
                   -> C.constraints * 'a
                   -> ('a,'ans) sc * ('a,'ans) fc
                   -> ans

     decide pfStep (G,D,C) (c,x) (sc,fc) = b

     Invariants
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses that *may not be left asynchronous*
       C is the goal that *may not be right asynchronus*
       c is not zero

     Decide on a formula to focus on.
       + First try linear hypotheses
       + Then try unrestricted hypotheses
  *)
  fun decide pfStep (G, D, C) (c,x) (sc,fc) =
      let
        (* decideLBang : unrestrictedCxt -> 'a -> 'ans
           decideLBang (Gb) x = b

           Gb is the subset of G that has not been focused on
        
           Try focusing on each unrestricted hypothesis.  If we
           fail, then call the failure continuation.

           *** Place for parallelism -- OR-parallelism
        *)
        fun decideLBang (Cxt.Empty) x = fc x
          | decideLBang (Cxt.Ext (G',A)) x =
            (case C of
               Prop.Atom _ =>
                 (case headcheck (G,D,A,C) c of
                    NONE => decideLBang G' x
                  | SOME (c',gs) =>
                      let
                        fun resl d (nil) = sc d
			  | resl (d as (cx',fc)) (h::t) =
                              pfStep h cx' (fn d' => resl d' t, fc)
		      in
                        resl ((c',x), fn x' => decideLBang G' x') gs
		      end)
             | _ => pfStep (F.LeftFocus (G, D, A, C)) (c,x)
                           (sc, fn x' => decideLBang G' x'))

        (* decideL : C.linearCxt * C.linearCxt -> 'a -> 'ans
           decideL (Db,Df) x = b

           Db is the subset of D that has not been been focused on
           Df is the subset of D not including the Db
        
           Try focusing on each linear hypothesis.  If we fail,
           then try focusing on unrestricted hypotheses.

           *** Place for parallelism -- OR-parallelism
        *)
        fun decideL (Cxt.Empty, _) x = decideLBang G x
          | decideL (Cxt.Ext (Db, (A,e)), Df) x =
            let
              (* skip : 'a -> 'ans *)
              fun skip x' = decideL (Db, Df ++ (A,e)) x'

              (* apply : C.constraints -> 'ans *)
              fun apply c =
                  (case C of
		     Prop.Atom _ =>
                       (case headcheck (G, Cxt.combine (Db,Df),A,C) c of
			  NONE => skip x
			| SOME (c',gs) =>
			  let
                            fun resl d (nil) = sc d
			      | resl (d as (cx',fc)) (h::t) =
                                  pfStep h cx' (fn d' => resl d' t, fc)
			  in
                            resl ((c',x),skip) gs
			  end)
		   | _ => (* TODO Possibly use a more efficient data
                             structure for linear contexts her *)
                          pfStep (F.LeftFocus (G, Cxt.combine (Db,Df), A, C))
                                 (c,x) (sc, skip))
            in
              if C.bexpEq (e,C.zero) then skip x
              else if C.bexpEq (e,C.one) then apply c
              else 
                let
                  val c' = C.addConstraints (c, C.constrainExp (e,true))
                in
                  if C.unsatisfiable c' then skip x
                  else apply c'
                end
            end
      in
        case C of
          (* C is atomic -- must try to focus on a hypothesis *)
          Prop.Atom P => decideL (D, Cxt.Empty) x
        | (* C is not atomic -- try to focus on the right first *)
          C => pfStep (F.RightFocus (G, D, C)) (c,x)
                      (sc, fn x' => decideL (D, Cxt.Empty) x')
      end


  (* pfStep : F.sequent
              -> C.constraints * 'a
              -> 'a F.sc * 'a F.fc
              -> F.ans

     pfStep seq (c,x) (sc,fc) = b

     Make a step in the proof search.
  *)
  fun pfStep seq d k =
      let
        fun doit (F.RightAsync (s)) = F.rightAsync pfStep s d k
          | doit (F.LeftAsync  (s)) = F.leftAsync pfStep s d k
          | doit (F.Decide     (s)) =
            if Constants.USE_HEADCHECK then decide pfStep s d k
	    else F.decide pfStep s d k
          | doit (F.RightFocus (s)) = F.rightFocus pfStep s d k
          | doit (F.LeftFocus  (s)) = F.leftFocus pfStep s d k

        val _ = dbgMsg ("Proving " ^ (F.sequentToString seq))
      in
        doit seq
      end


  (*********************)
  (* Proving Interface *)
  (*********************)

  (* prove : Prop.prop Cxt * Prop.prop Cxt * Prop.prop -> bool *)
  fun prove (G, D, C) =
      let
        val _ = F.C.init ()

        fun mkInitCxt (Cxt.Empty) = Cxt.Empty
          | mkInitCxt (Cxt.Ext (D', A)) = Cxt.Ext (mkInitCxt D', (A,F.C.one))
      in
        pfStep (F.RightAsync (G, Cxt.Empty, mkInitCxt D, C)) (F.C.empty,())
               (fn _ => true, fn _ => false)
      end

end; (* functor LLProver *)
