(************************************************************************)
(* obdd.sml                                                             *)
(*                                                                      *)
(* Description:                                                         *)
(*    An implementation of ordered binary decision diagrams (OBDDs).    *)
(*                                                                      *)
(* Author:                                                              *)
(*    Evan Chang                                                        *)
(************************************************************************)

(* $Id: *)

(**************************************************************)
(* structure Obdd                                             *)
(*                                                            *)
(* An OBDD is represented by tables mapping nodeids to the    *)
(* corresponding node and an inverse mapping from  nodes to   *)
(* nodeids.                                                   *)
(*                                                            *)
(* Node IDs are 0,1,... with 0 and 1 reserved for the "0"     *)
(* and "1" nodes.                                             *)
(*                                                            *)
(* In this implementation, each BDD has its own table         *)
(* rather than a global unique table to be able to have       *)
(* each obdd independent of any other.  This is at the        *)
(* cost of memory because we cannot share nodes across        *)
(* BDDs.  This also means that we cannot cache computations   *)
(* of apply across calls.                                     *)
(**************************************************************)
structure Obdd :> OBDD =
struct

  (******************************************)
  (* Boolean variables                      *)
  (******************************************)

  (* Variables are represented by integers *)
  type var = int

  (* Node IDs *)
  type nodeid  = int

  (* Node *)
  datatype node =
    Zero
  | One
  | Node of var * nodeid * nodeid 


  (******************************************)
  (* Binary boolean operator                *)
  (******************************************)
  datatype oper =
    And
  | Or
  | Imp
  | BiImp
  | Xor
  | Nand
  | Nor

  fun compl 0 = 1
    | compl 1 = 0
    | compl _ = raise Fail "invalid boolean value"

  fun opr And   (x,y) = if x=1 andalso y=1 then 1 else 0
    | opr Or    (x,y) = if x=1 orelse y=1 then 1 else 0
    | opr Imp   (x,y) = if x=1 andalso y=0 then 0 else 1
    | opr BiImp (x,y) = compl (opr Xor (x,y))
    | opr Xor   (x,y) = (x + y) mod 2
    | opr Nand  (x,y) = compl (opr And (x,y))
    | opr Nor   (x,y) = compl (opr Or (x,y))


  (******************************************)
  (* Tables                                 *)
  (******************************************)
  val BIG_PRIME = 1073741789
  fun hashPair (i,j) = ((( ((i + j)*(i + j + 1)) mod BIG_PRIME) div 2) + i)
                       mod BIG_PRIME

  structure HKey :> HASHABLE where type key = node =
  struct
    type key = node
    type hash = int

    fun hash (Zero) = 0
      | hash (One) = 1
      | hash (Node (v,l,h)) =  hashPair (v, hashPair (l,h))
    fun index x = x
    val eq = op=
  end (* structure Obdd.HKey *)

  structure TKey :> HASHABLE where type key = nodeid =
  struct
    type key = nodeid
    type hash = int

    fun hash x = x
    fun index x = x
    val eq = op=
  end (* structure Obdd.TKey *)

  structure GKey :> HASHABLE where type key = nodeid * nodeid =
  struct
    type key = nodeid * nodeid
    type hash = int

    fun hash x = hashPair x
    fun index x = x
    fun eq ((x1,y1),(x2,y2)) = x1=x2 andalso y1=y2
  end (* structure Obdd.GKey *)

  structure H
  :> HASHMAP where type key = HKey.key = HashMapFn(structure K = HKey)

  structure T
  :> HASHMAP where type key = TKey.key = HashMapFn(structure K = TKey)

  structure G
  :> HASHMAP where type key = GKey.key = HashMapFn(structure K = GKey)

  structure UT =
  struct

    type uniquetable = int ref          (* next available id *)
                       * nodeid H.table (* (var,low,high) |-> nodeid *)
                       * node T.table   (* nodeid |-> (var,low,high) *)

    (* init : unit -> uniquetable *)
    fun init () =
        let
          val h : nodeid H.table = H.new 10
          val t : node T.table   = T.new 10
          val _ = T.insert t (0, Zero)
          val _ = T.insert t (1, One)
	in
          (ref 2, h, t)
	end

    (* mk : uniquetable -> var * nodeid * nodeid -> nodeid
       mk ut (var,low,high) = nid

       Search to see if Node(var,low,high) exists.  If it does,
       then return its nodeid.  Otherwise, create a new node
       and return its nodeid.
    *)
    fun mk (rnid as ref nextid, h, t) (n as (_, low, high)) =
        if low=high then
          low
        else 
          (case H.lookup h (Node n) of
             NONE    => let
                          val _ = H.insert h (Node n, nextid)
                          val _ = T.insert t (nextid, Node n)
                          val _ = rnid := nextid + 1
	                in
                          nextid
	                end
           | SOME id => id)

  end (* structure Obdd.UniqueTable *)


  (******************************************)
  (* OBDD                                   *)
  (******************************************)
  type obdd = nodeid           (* the root node *)
              * UT.uniquetable (* the node mapping tables *)

  (* zero : obdd *)
  val zero = (0, UT.init ())

  (* one : obdd *)
  val one  = (1, UT.init ())

  (* mkVar : var -> obdd *)
  fun mkVar v = 
      let
        val ut = UT.init ()
        val id = UT.mk ut (v,0,1)
      in
        (id, ut)
      end

 (* apply : oper -> obdd * obdd -> obdd
    apply oper (bdd1,bdd2) = bdd

    Create the bdd that represents "*bdd1* oper *bdd2*" where *x* means
    the boolean expression that is represented by x.
 *)
 fun apply oper ((r1,(ref nextid1,_,t1)), (r2,(ref nextid2,_,t2))) =
     let
       val ut = UT.init ()
       val cache : nodeid G.table = G.new (nextid1 * nextid2)

       (* compareNodes : node * node -> order *)
       fun compareNodes (Node (v1,_,_), Node (v2,_,_)) = Int.compare (v1,v2)
         | compareNodes (n1,n2) =
           let
             (* leaf : node -> bool *)
	     fun leaf (Zero) = true
               | leaf (One) = true
               | leaf _ = false

             exception Impossible
	   in
             if leaf n1 andalso leaf n2 then EQUAL  (* both are leaves *)
             else if leaf n1 then GREATER           (* n1 is leaf, n2 is not *)
             else if leaf n2 then LESS              (* n2 is leaf, n1 is not *)
             else raise Impossible                  (* both are terminals *)
	   end
       
       (* apply' : nodeid * nodeid -> nodeid *)
       fun apply' (l,r) =
           let
             val $-> = G.lookup cache
             val $<- = G.insert cache

             (* leaf : nodeid -> bool *)
             fun leaf nid = nid=0 orelse nid=1
	   in
             case $->(l,r) of
               SOME a => a
             | NONE =>
                 (* Base Case *)
                 if leaf l andalso leaf r then
                   let
		     val b = opr oper (l,r)
                     val _ = $<-((l,r),b)
		   in
                     b
		   end
                 (* Inductive Cases *)
                 else
                   let
                     val nl = valOf (T.lookup t1 l)
		     val nr = valOf (T.lookup t2 r)

		     fun vlh (Zero) = (~1,0,0)
                       | vlh (One) = (~1,1,1)
		       | vlh (Node (v,l,h)) = (v,l,h)

                     val (varL,lowL,highL) = vlh nl
                     val (varR,lowR,highR) = vlh nr
                   in
                     case compareNodes (nl,nr) of
                       EQUAL =>
		         let
                           val low = apply' (lowL, lowR)
                           val high = apply' (highL, highR)
                           val b = UT.mk ut (varL, low, high)
                           val _ = $<-((l,r),b)
			 in
                           b
			 end
                     | LESS =>
		         let
                           val low = apply' (lowL, r)
                           val high = apply' (highL, r)
                           val b = UT.mk ut (varL, low, high)
                           val _ = $<-((l,r),b)
			 in
                           b
			 end
                     | GREATER =>
		         let
                           val low = apply' (l, lowR)
                           val high = apply' (l, highR)
                           val b = UT.mk ut (varR, low, high)
                           val _ = $<-((l,r),b)
			 in
                           b
			 end
		   end
	   end
     in
       (apply' (r1,r2), ut)
     end

  (* not : obdd -> obdd *)
  fun not bdd = apply Xor (bdd,one)

  (* eq : obdd * obdd -> bool *)
  fun eq (bdd1,bdd2) =
      let
        val (r,_) = apply BiImp (bdd1,bdd2)
      in
        1=r
      end

  (* toString : obdd -> string *)
  fun toString (root, (_, _, t)) =
      let
        val INDENT = 4

        (* padLeft : (int * string) -> string *)
        fun padLeft (0,s) = s
          | padLeft (n,s) = padLeft (n-1, " " ^ s)

        (* toString' : (nodeid * int) -> string *)
        fun toString' (p, depth) =
            let
              val n = valOf (T.lookup t p)
              val $ = Int.toString
	    in
              case n of
                Zero => padLeft (INDENT*depth, "(0)\n")
              | One  => padLeft (INDENT*depth, "(1)\n")
              | Node (v,l,r) => (toString' (r, depth+1))
                                ^ (padLeft (INDENT*depth,
                                            "(" ^ $p ^ ":x" ^ $v ^ ")\n"))
                                ^ (toString' (l, depth+1))
	    end
      in
        toString' (root, 0)
      end

end; (* structure Obdd *)
