(************************************************************************)
(* prop.sml                                                             *)
(*                                                                      *)
(* Description:                                                         *)
(*     Abstract syntax of linear logic propositons.                     *)
(*                                                                      *)
(* Author:                                                              *)
(*     Evan Chang                                                       *)
(************************************************************************)

(* $Id: prop.sml,v 1.2 2002/04/24 02:12:13 bechang_conductor Exp $ *)


structure Prop :> PROP =
struct

  datatype prop =                       (* A ::=         *)
    Atom of string                      (*       P       *)
  | Tensor of prop * prop               (*     | A (x) B *)
  | Lolli of prop * prop                (*     | A -o B  *)
  | Unit                                (*     | 1       *)
  | With of prop * prop                 (*     | A & B   *)
  | Or of prop * prop                   (*     | A (+) B *)
  | Zero                                (*     | 0       *)
  | Top                                 (*     | T       *)
  | Implies of prop * prop              (*     | A => B  *)
  | Bang of prop                        (*     | !A      *)

  fun eq (Atom P, Atom Q) = (P = Q)
    | eq (Tensor (A, B), Tensor (C, D)) = eq (A, C) andalso eq (B, D)
    | eq (Lolli (A, B), Lolli (C, D)) = eq (A, C) andalso eq (B, D)
    | eq (Unit, Unit) = true
    | eq (With (A, B), With (C, D)) = eq (A, C) andalso eq (B, D)
    | eq (Or (A, B), Or (C, D)) = eq (A, C) andalso eq (B, D)
    | eq (Zero, Zero) = true
    | eq (Top, Top) = true
    | eq (Implies (A, B), Implies (C, D)) = eq (A, C) andalso eq (B, D)
    | eq (Bang (A), Bang (B)) = eq (A, B)
    | eq _ = false


  (* precedence : prop -> int

     Assign precedence values to operators for printing.
  *)
  fun precedence (Atom _)    = 10
    | precedence (Unit)      = 10
    | precedence (Zero)      = 10
    | precedence (Top)       = 10
    | precedence (Bang _)    = 8
    | precedence (Tensor _)  = 6
    | precedence (With _)    = 6
    | precedence (Or _)      = 4
    | precedence (Lolli _)   = 2
    | precedence (Implies _) = 2

  fun toString C =
      let
        (* toString' : prop * int -> string
           toString' (C,n) = s

           n is the binding strength of the operator to the left
        *)
    fun toString' (C, n) =
            let
              fun parens C = "(" ^ toString' (C,0) ^ ")"
              val pr = precedence C
            in
              if n > pr then
                parens C
              else
                (case C of
                   Atom P        => P
                 | Tensor (A,B)  => toString' (A,pr) ^" (x) "^ toString' (B,pr)
                 | Lolli (A,B)   => toString' (A,pr) ^" -o "^ toString' (B,pr)
                 | Unit          => "1"
                 | With (A,B)    => toString' (A,pr) ^" & "^ toString' (B,pr)
                 | Or (A,B)      => toString' (A,pr) ^" (+) "^ toString' (B,pr)
                 | Zero          => "0"
                 | Top           => "T"
                 | Implies (A,B) => toString' (A,pr) ^" => "^ toString' (B,pr)
                 | Bang A        => "!" ^ toString' (A,pr))
            end
      in
        toString' (C,0)
      end

end; (* structure Prop *)
