(* $Id: tester.sml,v 1.7 2002/06/21 05:41:36 bechang_conductor Exp $ *)

(* Simple test harness for LLProver *)

(* Prover Structure *)
structure ProverSML = LLProver(structure Prop' = Prop
                               structure Cxt' = Cxt
                               structure Obdd' = Obdd)
structure ProverCML = LLProverCML(structure Prop' = Prop
                                  structure Cxt' = Cxt
                                  structure Obdd' = Obdd)
structure ProverCC = LLProverCC(structure Prop' = Prop
                                structure Cxt' = Cxt
                                structure Obdd' = Obdd)


(* Signature for specifying a test *)
signature LLPROVERTEST =
sig

  structure LLProver : LLPROVER

  (* A test is ((G,D,C),b) where
       G (Gamma) contains unrestricted hypotheses
       D (Delta) contains linear hypotheses
       C is the goal to be proven
       b is the expected result
  *)
  type test = (LLProver.Prop.prop LLProver.Cxt.Cxt *
               LLProver.Prop.prop LLProver.Cxt.Cxt * LLProver.Prop.prop) * bool

  val tests : test list

end; (* signature LLPROVERTEST *)


(* Simple Tests *)
functor SimpleTests(structure LLProver' : LLPROVER) :> LLPROVERTEST =
struct

  structure LLProver = LLProver'

  local
    open LLProver.Prop
    open LLProver.Cxt
  
    val $ = Atom
    val <*> = Tensor
    val --@ = Lolli
    val #! = Bang
    val & = With
  
    infix  4 <*>
	infix  4 &
    infixr 2 --@

  in

    type test = (prop Cxt * prop Cxt * prop) * bool

    (****************)
    (* Propositions *)
    (****************)
  
    (* A (x) A *)
    val AxA = $"A" <*> $"A"
  
    (* B (x) A *)
    val BxA = $"B" <*> $"A"

    (* A (x) B *)
    val AxB = $"A" <*> $"B"
  
    (* A (x) B -o B (x) A *)
    val p1 = AxB --@ BxA
  
    (* A -o A (x) A *)
    val p2 = $"A" --@ AxA
  
    (* !A -o A (x) A *)
    val p3 = #!($"A") --@ AxA
  
  
    (*********)
    (* Tests *)
    (*********)
  
    (* .;. |- A (x) B -o B (x) A *)
    val t1 : test = ((Empty, Empty, p1), true)
  
    (* .;. |- A -o A (x) A *)
    val t2 : test = ((Empty, Empty, p2), false)
  
    (* .;A |- A -o A (x) A *)
    val t3 : test = ((Empty, Ext (Empty, $"A"), p2), true)
  
    (* A;. |- A -o A (x) A *)
    val t4 : test = ((Ext (Empty, $"A"), Empty, p2), true)
  
    (* .;. |- !A -o A (x) A *)
    val t5 : test = ((Empty, Empty, p3), true)

    (* .; A, B, B -o A |- A (x) B (x) T *)
    val t6 : test = ((Empty, Ext(Ext(Ext(Empty,$"A"),$"B"),$"B" --@ $"A"),
                     AxB <*> Top), true)

    (* A, B -o A; B |- A (x) B *)
    val t7 : test = ((Ext(Ext(Empty, $"A"), $"B" --@ $"A"), Ext(Empty, $"B"),
                      AxB), true)

    (* A; .; A -o B & C |- C *)
    val t8 : test = ((Ext(Empty, $"A"), Ext(Empty, $"A" --@ $"B" & $"C"),
	             $"C"), true)
  
    (* .; A, B, B -o C |- C *)
    val t9 : test = ((Empty, Ext(Ext(Ext(Empty, $"A"), $"B"), $"B" --@ $"C"),
                      $"C"), false)

    val tests : test list = [ t1, t2, t3, t4, t5, t6, t7, t8, t9 ]

  end

end; (* functor SimpleTests *)


(* Blocks World

   Motivating example for linear logic as described in Frank's
   notes that describes a world where various blocks are stacked
   on a table with a robot arm that can pick up and put down the
   blocks one at a time.

   The state of the world is described using the following propositons:

      on(x,y)       block x is on block y
      tb(x)         block x is on the table
      holds(x)      robot arm holds block x
      empty         robot arm is empty
      clear(x)      the top of the block x is clear

  Then, we can describe the possible state transitions in this world
  as follows:

     geton : Forall x. Forall y. empty (x) clear(x) (x) on(x,y)
                                 -o holds(x) (x) clear(y)
     gettb : Forall x. empty (x) (x) clear(x) (x) tb(x) -o holds(x)
     puton : Forall x. Forall y. holds(x) (x) clear(y)
                                 -o empty (x) on(x,y) (x) clear(x)
     puttb : Forall x. holds(x) -o empty (x) tb(x) (x) clear(x)

  Since we currently do not support first-order logic, we will
  have atomic propositons for each combination of the blocks and
  the corresponding state transition rules above.
*) 
functor BlocksWorld(structure LLProver' : LLPROVER) :> LLPROVERTEST =
struct

  structure LLProver = LLProver'

  local
    val ++ = LLProver.Cxt.add
    infix ++

    open LLProver.Prop
    open LLProver.Cxt
  
    val $ = Atom
    val <*> = Tensor
    val --@ = Lolli
  
    infix  4 <*>
    infixr 2 --@

    (* Names/Prefixes for the atomic propositions *)
    val pon    = "on"
    val ptb    = "tb"
    val pholds = "holds"
    val pempty = "empty"
    val pclear = "clear"

    (* genRules : string list -> prop Cxt
       genRules bs = G

       Generate the state transition rules using the block names given
       in bs.
    *)
    fun genRules l =
        let
          fun outerloop (nil, G) = G
            | outerloop (x::t, G) =
              let

                fun innerloop (nil, G') = G'
                  | innerloop (y::t', G') =
                    if x=y then innerloop (t', G')
                    else 
                      let
                        val puton = $(pholds ^ x) <*> $(pclear ^ y)
                                    --@ $pempty <*> $(pon ^ x ^ y)
                                        <*> $(pclear ^ x)
                        val geton = $pempty <*> $(pclear ^ x)
                                    <*> $(pon ^ x ^ y)
                                    --@ $(pholds ^ x) <*> $(pclear ^ y)
                      in
                        innerloop (t', G' ++ geton ++ puton)
                      end

                val puttb = $(pholds ^ x)
                            --@ $pempty <*> $(ptb ^ x) <*> $(pclear ^ x)
                val gettb = $pempty <*> $(pclear ^ x) <*> $(ptb ^ x)
                            --@ $(pholds ^ x)
                val G' = innerloop (l, G ++ gettb ++ puttb)
              in
                outerloop (t, G')
              end
        in
          outerloop (l, Empty)
        end

  in

    type test = (prop Cxt * prop Cxt * prop) * bool

    (*********)
    (* Tests *)
    (*********)

    val blocks1 = ["a", "b"]
    val rules1  = genRules blocks1
    val init1   = Empty ++ $pempty ++ $(ptb ^ "a") ++ $(pon ^ "b" ^ "a")
                        ++ $(pclear ^ "b")
    val goal1   = $(pon ^ "a" ^ "b") <*> Top
    val t1      = ((rules1, init1, goal1), true)

    val blocks2 = ["a"]
    val rules2  = genRules blocks2
    val init2   = Empty ++ $pempty ++ $(ptb ^ "a") ++ $(pclear ^ "a")
    val goal2   = $(pholds ^ "a")
    val t2      = ((rules2, init2, goal2), true)

    val tests   = [ t2 ]

  end

end; (* functor BlocksWorld *)


(* Addition

   A simple example that checks if the numbers in the linear
   context add to the number given in the goal.

   To set up 
*) 
functor Addition(structure LLProver' : LLPROVER) :> LLPROVERTEST =
struct

  structure LLProver = LLProver'

  local
    val ++ = LLProver.Cxt.add
    infix ++

    open LLProver.Prop
    open LLProver.Cxt
  
    val $ = Atom
    val <*> = Tensor
    val --@ = Lolli
  
    infix  4 <*>
    infixr 2 --@

    val % = Int.toString

    (* for : 'a -> int -> int -> (int * 'a -> 'a) -> unit

       For loop from lo to hi, inclusive 
    *)
    fun for a lo hi f =
	if lo > hi then a else let val a' = f (lo,a) in for a' (lo+1) hi f end

    (* genRules : int * int -> prop Cxt
       genRules (lo,hi) = G
       
       Generate addition rules for numbers in the range [lo,hi].
    *)
    fun genRules (lo,hi) =
          for (Empty) lo hi (fn (i,G) => 
            for G i hi (fn (j,G) =>
              Ext(G, $(%i) <*> $(%j) --@ $(%(i+j)))))
  in

    type test = (prop Cxt * prop Cxt * prop) * bool

    (*********)
    (* Tests *)
    (*********)

    val rules1 = genRules (0,2)
    val init1 = Empty ++ $(%1) ++ $(%1) ++ $(%2) ++ $(%0)
    val goal1 = $(%4)
    val t1      = ((rules1, init1, goal1), true)
  
    val goal2 = $(%3)
    val t2      = ((rules1, init1, goal2), false)

    val tests   = [ t1, t2 ]

  end

end; (* functor Addition *)



(* Tester *)
functor Tester(structure Test' : LLPROVERTEST) =
struct

  structure Test = Test' 

  local
    open Test.LLProver.Prop
    open Test.LLProver.Cxt

    fun cxtToString (Empty) = "."
      | cxtToString (G) =
        let
          fun cxtToString' (Ext (Empty, A)) = toString A
            | cxtToString' (Ext (G',A)) = cxtToString' G' ^ ", " ^ toString A
        in
          cxtToString' G
        end

  in

    (* test : Test.test -> unit *)
    fun test (J as (G,D,C), ans) =
        let
          val res = Test.LLProver.prove J
  
          val _ = TextIO.print (cxtToString G ^ " ; " ^ cxtToString D ^
                            "  |-  " ^ Test.LLProver.Prop.toString C ^ "\n")
  
          val _ = if res=ans then print "    PASSED - "
                  else            print "    FAILED - "
  
          val _ = TextIO.print ("Result: " ^ Bool.toString res ^
                                "\tShould be: " ^ Bool.toString ans ^ "\n")
        in
          ()
        end
  
    (* runTests : unit -> unit *)
    fun runTests () = List.app test Test.tests

  end

end; (* functor Tester *)


(* Test Structures *)
structure SimpleTestsSML = Tester(structure Test' =
                                   SimpleTests(structure LLProver' = ProverSML))
structure SimpleTestsCML = Tester(structure Test' =
                                   SimpleTests(structure LLProver' = ProverCML))
structure SimpleTestsCC = Tester(structure Test' =
                                  SimpleTests(structure LLProver' = ProverCC))

structure BlocksWorldSML = Tester(structure Test' =
                                   BlocksWorld(structure LLProver' = ProverSML))
structure BlocksWorldCML = Tester(structure Test' = 
                                   BlocksWorld(structure LLProver' = ProverCML))
structure BlocksWorldCC = Tester(structure Test' =
                                  BlocksWorld(structure LLProver' = ProverCC))

structure AdditionSML = Tester(structure Test' =
                                   Addition(structure LLProver' = ProverSML))

structure AdditionCML = Tester(structure Test' =
                                   Addition(structure LLProver' = ProverCML))

structure AdditionCC = Tester(structure Test' =
                                  Addition(structure LLProver' = ProverCC))
